# -*- coding: utf-8 -*-
# stuff to handle cgroups

import os
import rollback
import tempfile
import time
import procinfo
import signal
import sys

from subprocess import Popen, PIPE
from logging import log

class CgroupError(Exception):
	def __init__(self, msg):
		Exception.__init__(self, msg)

def is_mounted():
	'''quite ugly hack'''
	'''deprecated - not required, cgroup can be mounted multiple times'''
	proc = Popen('mount | grep cgroup', shell = True, stdout = PIPE)
	output = proc.communicate()[0]
	return len(output.split('\n')) >= 2

fs_name = 'fracas'

def umount_rollback(path):
	log('Umounting cgroup filesystem...')
	cmd = 'umount %s' % fs_name
	proc = Popen(cmd, shell = True, stdout = PIPE, stderr = PIPE)
	proc.communicate()
	res = proc.returncode
	if res != 0:
		log('Could not umount!')
	os.rmdir(path)
	return None

def mount(roll = True):
	'''mount cgroup filesystem'''
	mount_path = tempfile.mkdtemp(dir='/tmp')
	cmd = 'mount -t cgroup -o cpuset,cpu %s %s' % (fs_name, mount_path)
	proc = Popen(cmd, shell = True, stdout = PIPE, stderr = PIPE)
	out, err = proc.communicate()
	res = proc.returncode
	if res != 0: # error
		err = err.strip()
		raise CgroupError('Could not mount cgroup filesystem: %s' % err)
	if roll:
		rollback.register(umount_rollback, [mount_path])  # remember to umount cgroup fs
	return mount_path

#
##
### Cgroup low-level routines
##
#

from os.path import join, isdir, exists
import os

class Cpusets:
	
	def __init__(self, roll = True):
		self.cpath = mount(roll)
		self.source = '/tmp/fracas_screamer.c' # source code
		self.binary = '/tmp/fracas_screamer_binary'  # program
		rollback.register(rollback.delete_file, [ self.source ])
		rollback.register(rollback.delete_file, [ self.binary ])

	def get_numbers(self, path):
		f = open(path, 'r')
		tasks = f.readlines()
		f.close()
		return map(lambda x: int(x.strip()), tasks)

	def delete_dirs(self, path, auto = True):
		'''removes path dir and all its recursive subdirectories'''
		'''auto - True if you want to delete this dir too'''
		for name in os.listdir(path):
			subdir = join(path, name)
			if isdir(subdir):
				self.delete_dirs(subdir, True)
			if name == 'tasks' and auto == True: # we need to copy everythin' to root set
				done = False
				while not done:
					tids = self.get_numbers(join(path, name))
					done = self.migrate_tasklist(tids)
		if auto:
			os.rmdir(path) # delete itself

	def clean_cgroup(self):
		'''cleans all groups in cgroups directory'''
		log('Cleaning cgroup directory...')
		self.delete_dirs(self.cpath, auto = False)

	def get_path(self, name, *args):
		args = list(args) + [ name ]
		return join(self.cpath, *args)

	def create_group(self, *args):
		'''creates group, subgroups must exist'''
		path = join(self.cpath, *args)
		os.mkdir(path)
	
	def get_mems(self, *args):
		path = self.get_path('cpuset.mems', *args)
		f = open(path, 'r')
		line = f.readline().strip()
		f.close()
		return line
		
	def assign_mems(self, mems, *args):
		path = self.get_path('cpuset.mems', *args)
		f = open(path, 'w')
		f.write(mems)
		f.close()

	def assign_cpus(self, cpus, *args):
		'''assings cpus to a group'''
		path = self.get_path('cpuset.cpus', *args)
		cpustring = ','.join(map(lambda x: str(x), cpus))
		f = open(path, 'w')
		f.write(cpustring)
		f.close()

	def move_task(self, pid, *args):
		'''moves process to a given group'''
		path = self.get_path('tasks', *args)
		# print path, pid
		f = open(path, 'w')
		f.write(str(pid))
		f.close()

	def move_task_ignore_error(self, pid, *args):
		'''moves process but ignores errors'''
		ignored = [ 3, 22 ]  # no such process, important task
		try:
			self.move_task(pid, *args)
		except IOError, e:
			if e.errno not in ignored:
				raise
		return None

	def list_tasks(self, *args):
		path = self.get_path('tasks', *args)
		return self.get_numbers(path)
	
	def migrate_tasks(self, group1, group2):
		done = False
		while not done:
			tasks = self.list_tasks(*group1)
			done = self.migrate_tasklist(tasks, *group2)
		return None

	def migrate_tasklist(self, tasklist, *group):
		'''Migrate all (possible) tasks from group1 to group2 in a "safe" manner'''
		errors = { 3 : 0, 22 : 0 } # no such process, important task
		N = len(tasklist)   # we've got only this tasks
		for tid in tasklist:
			try:
				self.move_task(tid, *group)
			except IOError, e:
				if e.errno in [3, 22]:
					errors[e.errno] += 1
				else:
					raise
		return N == errors[22]   # True, if all possible tasks were migrated
			
	def set_cpushare(self, share, *args):
		path = self.get_path('cpu.shares', *args)
		f = open(path, 'w')
		f.write(str(share))
		f.close()
	
	def get_cpushare(self, *args):
		path = self.get_path('cpu.shares', *args)
		f = open(path, 'r')
		line = f.readline()
		f.close()
		return int(line.strip())
		
	def screamer_rollback(self, pid):
		'''this simply kills the process and waits for it'''
		log('Waiting for screamer (%d)...' % pid)
		os.kill(pid, signal.SIGKILL)
		os.waitpid(pid, 0)
	
	def compile_screamer(self):
		f = open(self.source, 'w')
		f.write('int main() {for(;;);}')
		f.close()
		ret = os.system("gcc %s -o %s" % (self.source, self.binary))
		if ret != 0:
			raise CgroupError('Could not compile screamer.')
	
	def create_screamer(self, proc, share):
		'''creates screamer for a given processor'''
		log('Starting screamer on processor %d (priority = %d)' % (proc, share))
		name = 'screamer%02d' % proc
		pid = os.fork()
		if not pid: # child
			try:
				os.execv(self.binary, [])
			except Exception, e:
				print e
			finally:
				sys.exit(1)         # this SHOULDN'T happen!
		rollback.register(self.screamer_rollback, [pid])
		self.create_group(name)
		mems = self.get_mems()  # get global mem. banks
		self.assign_mems(mems, name)
		self.assign_cpus([proc], name)
		self.set_cpushare(share, name)
		self.move_task(pid, name)

	def setup(self, allocs):
		'''setup cgroups to reflect allocations'''
		# find really used processors
		procs = procinfo.getinfo()['procs']   # get processor info
		allocs = dict(allocs)   # convert to a mapping
		cpus = map(lambda x: x, allocs)
		used = filter(lambda x: allocs[x] > 0, allocs)
		unused = list(set(cpus) - set(used))
		self.compile_screamer()
		# let's create one big group, and let's name it 'all'
		self.clean_cgroup()
		rollback.register(self.clean_cgroup, [])
		self.create_group('all')
		mems = self.get_mems()
		self.assign_mems(mems, 'all')  # whole memory
		self.assign_cpus(used, 'all')  # only USED processors
		# what I do here is an ugly hack
		# I set cpushare to something big and read what was set by the kernel
		# this should give me a maximum cpu share possible
		self.set_cpushare(1000000000, 'all')
		maxpr = self.get_cpushare('all')  # there is a limitation on the value, don't mess with the kernel
		log('Base priority is %d' % maxpr)
		self.migrate_tasks([], ['all'])
		# create "screamers"
		loads = {}
		for proc in used:
			espeed = float(allocs[proc])
			rspeed = float(procs[proc]['speed'])
			loads[proc] = 1. - espeed / rspeed
		maxload = max(loads.values())  # get maximum load
		base = maxpr * (1. - maxload)  # compute priority for 'all' group
		base = int(base)
		self.set_cpushare(int(base), 'all')
		log("'All' priority is %d" % base)
		for proc in loads:
			load = loads[proc]
			share = base * load / (1. - load)
			share = int(share)
			self.create_screamer(proc, share)
		return None
		