#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sys
import os
import signal
import procinfo
import freqinfo
import config
import allocation
import config
import rollback
import time
import cgroup
import errno
from logging import logger, log

class StartError(Exception):
	def __init__(self, msg):
		Exception.__init__(self, msg)

@logger('Reading configuration...')
def read_config():
	config.readconfig()

@logger('Reading CPU information...')
def detect_cpu():
	info = procinfo.getinfo()  # CPUs info
	
	phys_n = info['phys_num']
	log_n  = len(info['procs'])
	log('Detected %d physical CPU(s) and %d logical CPU(s).' % (phys_n, log_n))

	throttled = freqinfo.get_throttled()
	scaled    = freqinfo.get_cpufreq()

	if len(throttled) > 0:
		log('Cores ', throttled, ' are throttled. Results may be biased.')
	if not config.allow_throttling and len(throttled) > 0:
		raise StartError('Configuration forbids continuation.')

	if len(scaled) > 0:
		log('Cores ', scaled, ' have cpufreq on. Results may be biased.')
	if not config.allow_cpufreq and len(scaled) > 0:
		raise StartError('Configuration forbids continuation.')

@logger('Disabling CPU frequency scaling...')
def disable_cpufreq():
	try:
		freqinfo.disable_cpufreq()   # disable CPU freq. scaling
	except IOError, e:
		if e.errno == errno.EACCES:  # permission denied
			raise StartError('Permission denied. Are you root?')
		raise

@logger('Allocating virtual cores...')
def allocate_cpu(vcores):
	try:
		allocs = allocation.allocate(vcores)
	except allocation.AllocationError, e:
		raise StartError(e)
	log('Allocs: ', allocs)
	return allocs

@logger('Preparing cgroup filesystem...')
def prepare_cgroups(allocs, roll = True):
	try:
		csets = cgroup.Cpusets(roll=roll)
		log('Cgroup filesystem mounted at %s.' % csets.cpath)
		csets.setup(allocs)
	except cgroup.CgroupError, e:
		raise StartError(e)

# this is program entry point

try:
	argv = sys.argv
	log('-= Fracas 1.0 =-')
	if os.getuid() != 0:
		print "You must be the root user in order to use Fracas."
		sys.exit(1)
	read_config()
	detect_cpu()
	# disable_cpufreq()
	# time.sleep(1)  # i think that cpufreq governor change may have a small inertion
	allocs = allocate_cpu(config.vcores)
	prepare_cgroups(allocs)
	log('Entering loop... (CTRL+C to leave)')
	if len(argv) >= 2:  # send notification
		pid = int(argv[1])
		log('Sending notification to PID %d...' % pid)
		os.kill(pid, signal.SIGUSR1)
	while True:
		time.sleep(100)
except StartError, e:
	log('Error: ', e)
except config.ConfigError, e:
	log('Config error: ', e)
except KeyboardInterrupt:
	log('Leaving...')
finally:
	log('Rollback phase...')
	rollback.rollback()
	log('Finished.')
