# -*- coding: utf-8 -*-
# << Data mining >> classes

from collections import defaultdict
from utils import namedtuple
import pprint

class DataList(list):
	'''Kind of C# Linq for Python'''
	
	def __init__(self, labels, *args, **kwds):
		self._t = namedtuple('Row', labels)
		self._indices = set(xrange(len(labels)))
		self._labels = labels
		list.__init__(self, *args, **kwds)

	def addict(self, d):
		'''funny name!'''
		row = [ d[label] for label in self._labels ]
		self.append(tuple(row))

	def unlabel(self, label):
		'''converts label to index'''
		if isinstance(label, int):
			return label
		else:
			try:
				return self._labels.index(label)
			except ValueError:
				raise ValueError('unknown label: %s' % label)
	
	def unlabels(self, labels):
		'''converts labels to indices'''
		return [ self.unlabel(label) for label in labels ]
		
	def unlabels_both(self, labels):
		'''for given labels gives indices and indices for the rest'''
		labels = self.unlabels(labels)
		rest = sorted(self._indices - set(labels))
		return (labels, rest)
	
	def relabels(self, indices):
		'''converts indices to labels'''
		return [ self._labels[i] for i in indices ]
	
	def lgetter(self, indices):
		'''returns getter of specified subset of labels'''
		return lambda row: [ row[i] for i in indices ]
	
	def group(self, labels, order = None):
		'''Groups data using given labels
		   Returns a dictionary where keys are specified by labels
		   and values are lists of according rows '''
		ls, rs = self.unlabels_both(labels)
		rs_labels = self.relabels(rs)
		get_group = self.lgetter(ls)
		get_rest  = self.lgetter(rs)
		d = defaultdict(lambda: DataList(rs_labels))
		T = namedtuple('GroupRow', rs_labels)
		for row in self:
			group = tuple(get_group(row))
			d[group].append(T.make(get_rest(row)))
		if order:
			for key in d.keys():
				d[key].order(order)
		return d
	
	def order(self, labels):
		'''Sorts data using columns specified in labels'''
		ls = self.unlabels(labels)
		self.sort(key = self.lgetter(ls))
	
	def groupone(self, label, **kwds):
		'''Does the same thing as group() but just for one column'''
		d2 = self.group([label], **kwds)
		d = {}
		for key, value in d2.iteritems():
			d[key[0]] = value
		return d
	
	def itergroups(self, labels, **kwds):
		'''Return iterator over items of result of self.group()'''
		d = self.group(labels, **kwds)
		return d.iteritems()

	def gmap(self, labels, names, f):
		'''Groups data by everything apart from specified labels.
		   Then applies f to a list of grouped data'''
		ls, rs = self.unlabels_both(labels)
		labels = self.relabels(ls)
		ns = self.relabels(rs) + names  # new labels
		d = self.group(rs)  # group by everything else
		newdata = DataList(ns)
		for rest, x in d.iteritems():
			v = f(x)
			newrow = tuple(list(rest) + list(v))
			newdata.append(newrow)
		return newdata

	def select(self, labels, f = None):
		'''Selects (SQL alike) rows'''
		ls = self.unlabels(labels)
		labels = self.relabels(ls)
		result = DataList(labels)
		get_row = self.lgetter(ls)
		T = namedtuple('SelectRow', labels)
		for row in self:
			row = get_row(row)
			if f:
				row = f(T.make(row))
			result.append(row)
		return result

	def filter(self, f):
		'''Filters data according to a given predicate'''
		newdata = DataList(self._labels)
		for row in self:
			if f(self._t.make(row)):   # name tuple
				newdata.append(row)
		return newdata

	def row(self, i):
		'''direct getitem method'''
		return list.__getitem__(self, i)
	
	def __getitem__(self, i):
		'''overloaded item getter => returns named tuple'''
		return self._t.make(self.row(i))
	
	def unique(self):
		'''Returns a unique set of rows'''
		self.sort()
		newdata = DataList(self._labels)
		if len(self) > 0:
			newdata.append(self.row(0))
			for i in xrange(1, len(self)):
				row = self.row(i)
				if row != self.row(i-1):
					newdata.append(row)
		return newdata
	
	def zip(self, prefix = 'c'):
		'''zips all data, returns list of zips'''
		labels = [ prefix + str(c) for c in xrange(len(self)) ]
		return DataList(labels, zip(*self))
	
	def map(self, f):
		'''maps every row using specified function'''
		L = DataList(self._labels)
		for i in xrange(len(self)):
			row = f(self[i])
			L.append(row)
		return L

	def iterrows(self):
		for i in xrange(len(self)):
			yield self[i]

	def discard(self, labels):
		'''discards specified columns'''
		ls, rs = self.unlabels_both(labels)
		return self.select(rs)

if __name__ == '__main__':
	labels = ['node', 'solution', 'test', 'freq', 'val', 'id']
	m = DataList(labels)
	m.append( ( 'node1', 'sol1', 'test1', 1.623, 5, 0) )
	m.append( ( 'node1', 'sol2', 'test1', 3.000, 18, 0) )
	m.append( ( 'node2', 'sol2', 'test1', 1.623, 25, 0 ) )
	m.append( ( 'node1', 'sol1', 'test1', 2.000, 7, 1) )
	m.append( ( 'node1', 'sol1', 'test1', 2.000, 8, 2 ) )
	m.append( ( 'node1', 'sol2', 'test1', 1.623, 15, 1) )
	m.append( ( 'node1', 'sol2', 'test1', 2.000, 29, 2 ) )
	m.append( ( 'node2', 'sol2', 'test1', 2.000, 27, 1 ) )
	m.append( ( 'node2', 'sol2', 'test1', 3.000, 28, 2 ) )
	m.append( ( 'node2', 'sol2', 'test3', 2.000, 37, 0 ) )
	m.append( ( 'node2', 'sol2', 'test3', 2.000, 39, 1 ) )
	m.append( ( 'node2', 'sol2', 'test3', 3.000, 38, 2 ) )
	pprint.pprint( m.group(['node', 'solution']) )
	for (node, solution), rest in m.itergroups(['node', 'solution']):
		print 'Graph for: ', node, solution
		print rest
		
	def f(L):
		s = ''
		for val, id in L:
			s += ',' + str(val) + ':' + str(id)
		return (s,)
	
	print m.gmap(['val', 'id'], ['val'], f)
	print m.select(['node', 'test']).unique()
	print m.groupone('node')
	print
	m2 = m.zip()
	print m2
	
	m.order(['freq'])
	
	print m.filter(lambda x: x.node == 'node1')  # thanks to do namedtuple this works nicely
	print m.map(lambda row: row.replace(id = 0)).map(lambda x: x.replace(id = 1))
	