#!/usr/bin/env python
# -*- coding: utf-8 -*-
# post processing of the results
# fork of previous framework
# this is targeted to Python 2.6
# however with some work (mostly mining.py!) it should work on version 2.5

from __future__ import with_statement 

import re
import glob
import sys
import os
import os.path
import shutil
from os.path import join

import hashfile
import simplejson as js
from mining import DataList

TITLES = { 
			'fracas' : 'Fracas', 
			'cpulim1' : 'CPU-Lim1',
			'cpulim2' : 'CPULim1-ng',
			'cpulim3' : 'CPU-Lim2', 
			'cpufreq' : 'CPU-Freq',
			
			'stressint' : 'StressInt',
			'linpack'   : 'Linpack',
			'sleeper'   : 'Sleeper',
			'stream'    : 'Stream',
			'threads'   : 'Threads',
			'udper'     : 'UDPer',
			'cpudper'   : 'CPUDPer',
			'procs'     : 'Processes',
		}

UNITS = {
			'stressint' : 'Loops / s',
			'linpack'   : 'GFLOP / s',
			'sleeper'   : 'Loops / s',
			'stream'    : 'GB / s',
			'threads'   : 'Loops / s',
			'udper'     : 'Sends / s',
			'cpudper'   : 'Sends / s',
			'procs'     : 'Loops / s',
		}

def human_bench(bench):
	return TITLES.get(bench, bench)

def get_filelist(inputdir):
	'''gets list of files with results'''
	flist_path = join(inputdir, '!files')
	with open(flist_path, 'r') as f:
		return [ name.strip() for name in f ]

def map_result(res):
	'''extract important info from the results'''
	bench = res['bench']
	varss = res['vars']
	res[u'result'] = varss[bench]
	res[u'speeds'] = tuple(res[u'speeds'])

def assert_consistency(results):
	'''asserts the same set of variables in every result'''
	if len(results) == 0:
		raise AssertionError('No results!')
	keys = set(results[0])
	for r in results:
		if set(r.keys()) != keys:
			raise AssertionError('Inconsistent results!')
	return sorted(keys)

def to_datalist(results):
	'''convert data to datalist'''
	labels = assert_consistency(results)
	data = DataList(labels)
	for r in results:
		data.addict(r)
	return data

# postprocessing + other

def average_tstudent(L):
	'''computes 95% confidence interval for the given data'''
	from statistics import mean, tinterval
	m = mean(L)
	a, b, dx = tinterval(0.95, L)
	return (m, dx)

def average_results(data, result = 'result', id = 'id', error = 'error', count = None):
	'''averages results and computes confidence intervals'''
	def confidence_int(rows):
		if count:
			assert len(rows) == count, 'The average count is %d but should be %d' % (len(rows), count)
		return average_tstudent( [ row.result for row in rows ] )
	return data.gmap([result, id], [result, error], confidence_int)

def read_results(inputdirs):
	'''read results' data'''
	results = []
	for inputdir in inputdirs:
		for fname in get_filelist(inputdir):
			fname = join(inputdir, fname)
			d = hashfile.loadfile(fname)      # load hashfile format
			result = js.loads(d['test'])  # extract json result info
			result['inputdir'] = inputdir
			map_result(result)
			results.append(result)        # add to the results' list
	data = to_datalist(results)
	return data

def human_units(unit):
	return UNITS.get(unit, unit)

def human_title(title):
	return TITLES.get(title, title)

def human_series(series):
	'''Sorted, human-friendly labels'''
	hlabels = [ (human_title(l), l) for l in series ]
	for hlabel, label in sorted(hlabels):
		yield (hlabel, series[label])

def plot(series, xlim = None, ylim = None, markersize = 4, aspect = 170. / 285., margins = {}):
	'''Plots specified series'''
	from pylab import rc, figure, clf, axes, errorbar, legend, plt, grid
	scale = 1.5
	twidth = 2 * 2.913385 * scale # springer
	twidth *= 0.95     # occupy 95%
	w = twidth * 0.5
	h = aspect * w
	markersize *= scale
	rc('font', size = scale * 6)
	rc('axes', labelsize = scale * 11)
	rc('legend', fontsize = scale * 6.3)
	rc('xtick', labelsize = scale * 6.3)
	rc('ytick', labelsize = scale * 6.3)
	margs = { 'left' : 0.16, 'right' : 0.93, 'top' : 0.93, 'bottom' : 0.16 }
	for m in margs:
		if m not in margins:
			margins[m] = margs[m]
	rc('figure.subplot', **margins)
	figure(figsize = (w,h) )
	grid(True)
	if xlim or ylim: axes().set_autoscale_on(False)
	if xlim: axes().set_xlim( xlim )
	if ylim: axes().set_ylim( ylim )
	# sf = FF(lambda x, y: '%d' % int(x))
	# plt.gca().yaxis.set_major_formatter(sf)
	colors  = ['b', 'g', 'r', 'c', 'm', 'y', 'k', 'w' ]
	markers = ['o', 'v', '^', 's', 'p' ]
	lines   = [ [1,1], [2,2], [2,5], [5,5], [5, 3, 2, 3] ]
	cid = 0
	for label, serie in human_series(series):
		xaxis, yaxis, yerrs = serie
		L = errorbar(xaxis, yaxis, yerr = yerrs, color = colors[cid], ls = '', 
		       label = label, marker=markers[cid], markersize=markersize, dashes = lines[cid])
		cid += 1
	lg = legend(fancybox=True, loc=0)
	return plt

def save_series(series, path, title, xlabel, ylabel, xlim = None, ylim = None, margins = {}):
	'''graphs given series to a file'''
	p = plot(series, xlim = xlim, ylim = ylim, margins = margins)
	p.xlabel(xlabel, size = 'large')
	p.ylabel(human_units(ylabel), size = 'large')
	if title:
		p.title(human_title(title))
	p.savefig(path)

def dump_series(series, path):
	'''Dump raw x,y data (for gnuplot?)'''
	f = open(path, 'w')
	for label, serie in human_series(series):
		xaxis, yaxis, yerrs = serie
		f.write('# %s\n' % label)
		for x,y,z in zip(xaxis, yaxis, yerrs):
			f.write('%.6f %.6f %.6f\n' % (x,y,z))
		f.write('\n\n')
	f.close()

def plot_series(series, path, title, xlabel, ylabel):
	'''Gnuplot it'''
	dat = path + '.dat'
	dump_series(series, dat)
	f = open(path, 'w')
	f.write('set term postscript\n')
	f.write('set output "figure.eps"\n')
	plots = []
	ind = 0
	for label, _ in human_series(series):
		plots.append("'%s' index %d title '%s' with lines" % (dat, ind, label))
		ind += 1
	f.write('plot ' + ', '.join(plots))

def group_series(data, **kwds):
	xdat = kwds['x']
	ydat = kwds['y']
	yerr = kwds.get('yerr', None)
	xerr = kwds.get('xerr', None)
	series = kwds['series']
	group = kwds['group']
	for gr, rest in data.groupone(group).iteritems():    # group by a given label
		graph_data = {}
		for serie, xy in rest.groupone(series).iteritems():   # group by a serie
			xy = xy.select([xdat, ydat, yerr])
			xy.order([xdat])    # sort by X axis
			x, y, e = xy.zip()
			graph_data[serie] = (x, y, e)
		yield (gr, graph_data)

def compute_ylimits(res, col, result = 'result', error = 'error', margin = 0.1):
	'''computes bounds on the results'''
	tres = res.select([col, result, error])
	lims = {}
	for val, rest in tres.groupone(col).iteritems():
		up = lambda x: x.result + x.error
		do = lambda x: x.result - x.error
		M = max(rest, key = up)
		m = min(rest, key = do)
		M = up(M); m = do(m); s = M - m; marg = s * margin
		lims[val] = (m - marg, M + marg)
		# print '%s results are in (%.4f, %.4f)' % (bench, m, M)
	return lims

def relate_series(series, name):
	'''recomputes series relatively to the specified dataset'''
	bx, by, be = series[name]
	nseries = {}
	for label in series:
		s = series[label]
		xs = s[0]; ys = list(s[1]); es = list(s[2])
		for i in xrange(len(ys)):
			ys[i] /= by[i]
			es[i] /= by[i]
		nseries[label] = (xs, tuple(ys), tuple(es) )
	return nseries

def ask(msg):
	yes = [ 'yes', 'y']
	no  = [ 'no', 'n' ]
	answer = None
	while answer not in (yes + no):
		answer = raw_input(msg).lower()
	return answer in yes

if __name__ == '__main__':
	argv = sys.argv
	if len(argv) < 4:
		print 'Syntax: python %s <test-suite> <output-dir> <results-dirs> ' % argv[0]
		sys.exit(1)
	suitefile = sys.argv[1]
	outputdir = sys.argv[2]
	inputdirs = sys.argv[3:]
	print 'Processing according to: %s' % suitefile
	if os.path.exists(outputdir):
		msg = 'Directory %s exists! Are you sure? ' % outputdir
		if ask(msg):
			shutil.rmtree(outputdir)
		else:
			sys.exit(1)
	os.makedirs(outputdir)
	results = read_results(inputdirs)
	# execute postprocess from the suite
	get_results = lambda: results
	get_output = lambda f: join(outputdir, f)
	suite = { 'get_results' : get_results, 'get_output' : get_output }
	execfile(suitefile, suite)
	postprocess = suite['postprocess']
	postprocess()  # invoke defined function
