# -*- coding: utf-8 -*-
# This file defines methods of testing.
# Benchmark class is used to define a specific test (or benchmark).
# Engine class is used to define a specific framework to-be-tested.
# You can find specific examples below.

import os
import subprocess
import time
import signal
import sys
import shutil
import datetime
import tempfile
from os.path import join, abspath

argv = sys.argv
Popen = subprocess.Popen
PIPE = subprocess.PIPE
MODCPU = '../bin/modcpu'
MODSCHED = '../bin/modsched'
MODRELAX = '../bin/modrelax'
RESTRICT = '../fracas/restrict.py'
FRACAS_CLEAN = '../fracas/clean.py'
FRACAS = '../fracas/fracas.py'
RUNTEST = '../tests/runtest'

def S(path):
	print 'Running: %s' % path
	res = os.system(path)
	assert res == 0, '"%s" returned non-zero code' % (path,)

def Sout(path):
	print 'Running: %s' % path
	p = Popen([ path ], shell = True, stdout=PIPE)
	res = p.communicate()[0]
	assert p.returncode == 0, '"%s" returned non-zero code' % (path,)
	return res

def Sall(path):
	print 'Running: %s' % path
	p = Popen([ path ], shell = True, stdout=PIPE, stderr=PIPE)
	res = p.communicate()
	assert p.returncode == 0, '"%s" returned non-zero code' % (path,)
	return res

def bashlist(path, f = lambda x: x):
	return map(f, Sout(path).split())

def get_dir():
	return os.path.dirname(os.path.abspath(sys.argv[0]))

def get_host():
	from platform import node
	return node()

def clean_cg():
	S('python %s' % FRACAS_CLEAN)

def set_freq(core, val):
	S(MODCPU + ' -c %d speed %s' % (core,str(val)))

def get_cpu_max(core):
	return Sout(MODCPU + ' -q -c %d cpu_max' % core)

def get_cpu_min(core):
	return Sout(MODCPU + ' -q -c %d cpu_min' % core)

def set_governor(core, name):
	m = get_cpu_min(core)
	M = get_cpu_max(core)
	S(MODCPU + ' governor %s' % name)  # set all governors to userspace
	S(MODCPU + ' min %s' % m)
	S(MODCPU + ' max %s' % M)

def restrict_cpus(cpus):
	cpus = ' '.join([ str(cpu) for cpu in cpus ])
	S('python %s %s' % (RESTRICT, cpus) )

def set_latency(val):
	S('%s latency %s' % (MODSCHED, val))

def set_granularity(val):
	S('%s min_granularity %s' % (MODSCHED, val))

def set_relaxing(val):
	S('%s %s' % (MODRELAX, val))

def get_my_path(filename):
	global MYDIR
	return join(MYDIR, filename)

class Benchmark:
	def __init__(self, name):
		self.name = name

	def run(self, cmd):
		cmdline = '%s %s' % (RUNTEST, self.name)
		cmdline = '%s "%s"' % (cmd, cmdline)
		out, err = Sall(cmdline)
		return out, err

class Engine:
	
	def __init__(self, test):
		self.test = test
		self.envs = {
			'latency' : self.set_latency,
			'granularity' : self.set_granularity,
			'speeds'  : self.set_speeds,
			'relaxing' : self.set_relaxing,
			}
	
	def nothing(self, value):
		pass
	
	def cmd(self):  # add command to run tests
		return join(get_dir(), '../bin/run')
	
	def set_relaxing(self, value):
		set_relaxing(value)
	
	def set_latency(self, value):
		set_latency(value)
	
	def set_granularity(self, value):
		set_granularity(value)
	
	def set_speeds(self, value):
		pass
	
	def reset(self):
		speeds = self.test['speeds']
		for core in xrange(len(speeds)):  # set governor
			set_governor(core, 'userspace')
			set_freq(core, get_cpu_max(core))
		for name, value in self.test.items():
			f = self.envs.get(name, self.nothing)
			f(value)
	
	def run(self, bench):
		print 'Resetting...'
		self.reset()
		try:
			print 'Preparing...'
			self.pre()
			time.sleep(0.5)  # wait
			print 'Running...'
			out, err = bench.run( self.cmd() )
		except Exception, e:
			print 'Exception:', e
			raise
		finally:
			print 'Cleaning up...'
			self.post()
		return out, err
	
	def pre(self):
		pass
	
	def post(self):
		pass
	
	def nonzero(self, speeds):
		return [ x[0] for x in enumerate(speeds) if x[1] != 0 ]

class Fracas(Engine):
	
	def __init__(self, test):
		Engine.__init__(self, test)
		self.wd = tempfile.mkdtemp()
	
	def __del__(self):
		try:
			os.remove(join(self.wd, 'config'))          # remove generated config
			os.rmdir(self.wd)
		except Exception, e:
			print 'Exception:', e, '(ignoring)'
	
	def set_speeds(self, speeds):
		config = join(self.wd, 'config')
		conf = open(config, 'w')
		conf.write('[main]\n')
		conf.write('allow_throttling: yes\n')
		conf.write('allow_cpufreq: yes\n')
		vcores = ' '.join([ '%.4f' % (x / 1000.) for x in speeds ])
		conf.write('vcores: %s' % (vcores,))
		conf.close()
	
	def handler(self, num, frame):
		print 'Signal received.'
		self.started += 1
	
	def unset_handler(self):
		signal.signal(signal.SIGUSR1, signal.SIG_DFL)
	
	def set_handler(self):
		self.started = 0
		signal.signal(signal.SIGUSR1, self.handler)
	
	def pre(self):
		self.set_handler()
		self.fracas = Popen(['python', FRACAS, str(os.getpid()) ], cwd = self.wd)
		counter = 0.0
		while not self.started:
			counter += 0.1
			time.sleep(0.1)
			assert counter <= 5, 'Fracas has not started?!'
		self.unset_handler()
	
	def post(self):
		self.unset_handler()  # not a mistake -- pre() may throw an exception
		os.kill(self.fracas.pid, signal.SIGINT)   # stop fracas
		out = self.fracas.communicate()[0]   # wait for fracas
		assert self.fracas.returncode == 0, 'Fracas returned a non-zero code'

class CpuFreq(Engine):
	
	def set_speeds(self, speeds):
		cpuset = self.nonzero(speeds)  # nonzero cores
		print 'Restricting to %s' % cpuset
		restrict_cpus(cpuset)
		for core in cpuset:
			set_freq(core, speeds[core])
	
	def post(self):
		clean_cg()  # unrestrict cpus

class CpuLim(Engine):
	'''General CPULim'''
	
	def __init__(self, test):
		Engine.__init__(self, test)
		self.cleanup = lambda: None  # no cleanup by default
	
	def cmd(self):
		raise NotImplementedError('Not implemented!')
	
	def set_speeds(self, speeds):
		speed = set(speeds)
		if len(speed) == 1:  # one speed everywhere
			pass
		elif len(speed) == 2:  # possibly not used cores
			assert min(speed) == 0, 'CPULim does not support multiple frequency levels.'
			self.cleanup = clean_cg  # only clean when you have to
			cpuset = self.nonzero(speeds)
			restrict_cpus(cpuset)
		else:
			assert False, 'CPULim does not support such a configuration.'
		cpu_max = get_cpu_max(0)
		speed = max(speed)
		self.ratio = float(speed) / float(cpu_max)
	
	def post(self):
		self.cleanup()

class CpuLim1(CpuLim):
	
	def cmd(self):
		cpulim1 = join(get_dir(), '../cpulim1/cpulim1')
		return '%s %.4f' % (cpulim1, self.ratio)

class CpuLim3(CpuLim):
	
	def cmd(self):
		cpulim3 = join(get_dir(), '../cpulim3/cpulim3')
		return '%s -c %.4f' % (cpulim3, self.ratio)

def get_date():
	now = datetime.datetime.now()
	return now.strftime("%A %d/%m/%Y %H:%M:%S")

def format_secs(secs):
	secs = int(secs)
	ss = secs % 60; secs /= 60
	ms = secs % 60; secs /= 60
	hs = secs
	return '%02d hs %02d mins %02d secs' % (hs, ms, ss)

# definitions
MYDIR = get_dir()
MODCPU = get_my_path(MODCPU)
MODSCHED = get_my_path(MODSCHED)
MODRELAX = get_my_path(MODRELAX)
RESTRICT = get_my_path(RESTRICT)
FRACAS_CLEAN = get_my_path(FRACAS_CLEAN)
FRACAS = get_my_path(FRACAS)
RUNTEST = get_my_path(RUNTEST)

benchmarks = [
	Benchmark('stressint'),
	Benchmark('threads'),
	Benchmark('procs'),
	Benchmark('sleeper'),
	Benchmark('linpack'),
	Benchmark('udper'),
	Benchmark('cpudper'),
	Benchmark('stream'),
]

all_benchmarks = {}

for bench in benchmarks:
	all_benchmarks[bench.name] = bench

all_engines = {
	'cpufreq'  : CpuFreq,
	'fracas'   : Fracas,
	'cpulim1'  : CpuLim1, 
	'cpulim3'  : CpuLim3,
	}

# main stuff to run tests

def runtest(spec):
	bench  = all_benchmarks[spec['bench']]
	eclass = all_engines[spec['engine']]
	engine = eclass(spec)
	out, err = engine.run(bench)
	return out, err
