# -*- coding: utf-8 -*-
# This file defines abstract ObjectServer.
# It is a TCP server which serves a set of Python objects using pickle serialization.
# You can further specify how to react when connection is done, broken, etc.
# This is actually done in DistestServer class which uses ObjectServer to provide tests.

import traceback
import socket
import pickle
import sys
import select
import struct
import os
import time
import re
from os.path import join

from comm import PORT, Error, recv_obj, send_obj

import hashfile
import simplejson as js

POLLIN = select.POLLIN
TIMEOUT = 60

class BrokenError(Error):
	def __init__(self):
		Error.__init__(self, 'broken connection')

class ProtoError(Error):
	def __init__(self):
		Error.__init__(self, 'protocol error')

def default_log(msg):
	print msg

class ObjectServer:

	def __init__(self, port = PORT, log = default_log):
		global TIMEOUT
		self.clients = {}
		self.port = port
		self.log = log
		self.s = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
		self.s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
		self.timeout = TIMEOUT
		
	def addr(self, cl):
		info = self.clients[cl]
		return '%s:%d' % (info['addr'], info['port'])

	def poll(self):
		poll = select.poll()
		poll.register(self.s, POLLIN)
		for cl in self.clients:
			poll.register(cl, POLLIN)
		events = poll.poll(1000)
		ready = set([ event[0] for event in events ])
		if self.s.fileno() in ready:
			self.handle_new()
			ready.remove(self.s.fileno())
		socks = []
		for sock in self.clients:
			if sock.fileno() in ready:
				socks.append(sock)
		return set(socks)

	def handle_client(self, cl):
		hsize = struct.calcsize('!i')
		info = self.clients[cl]
		size = len(info['buf'])
		if size < hsize:
			bytes = cl.recv(hsize - size)
			if not bytes:
				raise BrokenError()
			info['buf'] += bytes
			return (None, False)
		else:
			header = info['buf'][:hsize]
			osize = struct.unpack('!i', header)[0]
			rsize = size - hsize
			bytes = cl.recv(osize - rsize)
			if not bytes:
				raise BrokenError()
			info['buf'] += bytes
			if len(info['buf']) == hsize + osize:  # we've got an object!
				obj = info['buf'][hsize:]
				try:
					obj = pickle.loads(obj)
				except Exception:
					raise ProtoError()
				info['buf'] = ''
				return (obj, True)
			else:
				return (None, False)
	
	def get_objs(self):
		ready = self.poll()
		objs = {}
		broken = {}
		disconn = []
		for cl in ready:
			info = self.clients[cl]
			info['time'] = time.time() # refresh timeout
			try:
				obj, ok = self.handle_client(cl)
				if ok:
					if obj == None:  # end of transmission
						disconn.append(cl)
					else:
						objs[cl] = obj
			except (socket.error, Error), e:
				broken[cl] = e
		return objs, disconn, broken
	
	def check_timeouts(self):
		now = time.time()
		for cl in self.clients.keys():
			info = self.clients[cl]
			if now - info['time'] > self.timeout:
				self.handle_broken(cl, 'timeouted')
	
	def set_timeout(self, value):
		self.timeout = value
	
	def serve(self):
		self.s.bind( ('', self.port) )
		self.s.listen(40)
		self.log('Serving at port %d...' % (self.port, ))
		while not self.finished():
			objs, disconn, broken = self.get_objs()
			for client, e in broken.items():
				self.handle_broken(client, e)
			for client in disconn:
				self.handle_finish(client)
			for client, obj in objs.items():
				self.handle_obj(client, obj)
			self.check_timeouts()
		self.log('Finished.')

	def finished(self):
		return False

	def handle_new(self):
		conn, (addr, port) = self.s.accept()
		self.log('%s:%d => connected' % (addr, port))
		hostname, _, _ = socket.gethostbyaddr(addr)
		self.log('%s:%d => resolved to %s' % (addr, port, hostname))
		addr = hostname
		info = { 'addr' : addr, 'port' : port, 'buf' : '', 'time' : time.time() }
		addinfo = self.new(conn)
		if addinfo == None:
			conn.close()
		else:
			info.update(addinfo)
			self.clients[conn] = info

	def handle_broken(self, cl, e):
		self.log('%s => %s' % (self.addr(cl), e))
		self.broken(cl)
		cl.close()
		del self.clients[cl]

	def handle_finish(self, cl):
		self.log('%s => closed connection' % (self.addr(cl),))
		self.finish(cl)
		cl.close()
		del self.clients[cl]

	def handle_obj(self, cl, obj):
		self.message(cl, obj)

	def new(self, cl):
		'''Executed on new connected socket'''
		return {}
	
	def broken(self, cl):
		'''Executed on broken connection'''
		pass
	
	def finish(self, cl):
		'''Executed on finished connection'''
		pass

	def message(self, cl, obj):
		'''Executed on object received'''
		pass

class DistestServer(ObjectServer):
	
	def __init__(self, tests, outdir, *args, **kwds):
		ObjectServer.__init__(self, *args, **kwds)
		self.n = len(tests)
		self.todo = list(tests)
		self.done = []
		self.outdir = outdir
		self.start = time.time()   # mark the start
	
	def new(self, cl):
		info = {}
		if len(self.todo) == 0:
			self.log('No more tests to perform.')
			send_obj(cl, None)
			return None
		else:
			test = self.todo.pop()
			info['test'] = test
			self.open_streams(info)
			send_obj(cl, dict(test))
		return info
	
	def open_streams(self, info):
		info['out'] = ''
		info['err'] = ''
	
	def close_streams(self, info):
		del info['out']
		del info['err']
	
	def broken(self, cl):
		info = self.clients[cl]
		self.close_streams(info)
		test = info['test']
		self.todo.append(test) # return to pool
	
	def add_file(self, filename):
		f = open(join(self.outdir, '!files'), 'a')
		f.write(filename + '\n')
		f.close()
	
	def extract_vars(self, s):
		varss = {}
		exp = re.compile('^([a-zA-Z_]+)=([\\.0-9]+).*')
		for line in s.split('\n'):
			m = exp.match(line)
			if m:
				varss[m.group(1)] = float(m.group(2))
		return varss
	
	def get_res_dict(self, info):
		d = { 'stdout': info['out'], 'stderr' : info['err'] }
		test = dict(info['test'])
		test['vars'] = self.extract_vars(info['out'])
		d['test'] = js.dumps(test, indent = 4)  # convert to json
		return d
		
	def finish(self, cl):
		info = self.clients[cl]
		test = info['test']
		outfile = join(self.outdir, test['output'])
		f = open(outfile, 'w')
		hashfile.save(f, self.get_res_dict(info))
		f.close()
		self.add_file(test['output'])
		self.done.append(test) # it's done
		self.close_streams(info)
		done = len(self.done); n = self.n
		span = time.time() - self.start
		total = (span * n) / float(done)
		remain = (total - span) / 60.
		self.log('Test \'%s\' is done by %s.' % (test['output'], self.addr(cl)))
		self.log('%d / %d tests done (around %d minutes more).' % (done, n, remain))
	
	def finished(self):
		return len(self.done) == self.n
	
	def message(self, cl, obj):
		info = self.clients[cl]
		if not self.isfine(obj):
			self.handle_broken(cl, 'wrong object (%r)' % (obj,))
		else:
			typ, bytes = obj
			if typ == 'log':
				self.log('%s => %s' % (self.addr(cl), bytes))
			elif typ == 'out':
				info['out'] += bytes
			elif typ == 'err':
				info['err'] += bytes

	def isfine(self, obj):
		ok = isinstance(obj, tuple) and len(obj) == 2 
		ok = ok and obj[0] in ['out','log','err'] and isinstance(obj[1], str)
		return ok
